/*********************************************************************

                    HS-15Pの簡易湿度計 for Arduino

本ソースリストおよびソフトウェアは、ライセンスフリーです。
個人での利用は自由に行えます。著作権表示の改変は禁止します。

                               Copyright (c) 2012 Wataru KUNINO
                               http://www.geocities.jp/bokunimowakaru/

再配布については下記の著作者の承諾を得る必要があります。
（本ソフトウェアの大半に下記の検討結果が含まれています）

		HS15湿度センサーと簡易回路による湿度計実験
		渡波 郁
		http://homepage3.nifty.com/sudamiyako/zk/hs15/hs15.htm
	
*********************************************************************/
/*
参考文献

本ソフトの開発のために下記の文献を参考にしました。

・HS15湿度センサーと簡易回路による湿度計実験
	渡波郁	http://homepage3.nifty.com/sudamiyako/zk/hs15/hs15.htm
・Arduinoをはじめよう (オライリージャパン)
	Massimo Banzi著 船田功訳
・Arduino 日本語リファレンス
	http://www.musashinodenpa.com/arduino/ref/index.php
*/
#include <SD.h>

#define	portSCK			19							// SCKポート番号(ANALOG 5)
#define	portDATA		18							// DATAポート番号(ANALOG 4)
#define	sensorVDD		5							// センサー電圧

byte HS15_P_OUT;
byte HS15_P_IN;

// SCKを"H"レベルに設定する関数
void SCK_H(void) {
    digitalWrite(portSCK, HIGH);
    delayMicroseconds(50);
}
// SCKを"L"レベルに設定する関数
void SCK_L(void) {
    digitalWrite(portSCK, LOW);
    delayMicroseconds(50);
}

// DATAを"H"レベルに設定する関数
void DATA_H(void) {
    digitalWrite(portDATA, HIGH);
    delayMicroseconds(50);
}
// DATAを"L"レベルに設定する関数
void DATA_L(void) {
    digitalWrite(portDATA, LOW);
    delayMicroseconds(50);
}

// SHTポートの初期化
void SHTInit(void) {
	pinMode(portSCK,OUTPUT);
	pinMode(portDATA,OUTPUT);
}

// SHT TSシーケンス
void SHTTSSeq(void) {
	pinMode(portSCK,OUTPUT);
	pinMode(portDATA,OUTPUT);
    DATA_H();    SCK_H();
    DATA_L();    SCK_L();    SCK_H();
    DATA_H();    SCK_L();
}

// バイト・データ送信
byte SHTWrite(byte dat) {
    byte i, rdat;

	pinMode(portDATA,OUTPUT);
    // DATAビットの出力
    for(i = 0; i < 8; i++) {
        SCK_L();
        if(dat & 0x80) {        // MSB
            // "1"
            DATA_H();
        } else {
            // "0"
            DATA_L();
        }
        SCK_H();
 
        dat <<= 1;              // 次のビット
    }
 
    SCK_L();
 

    // ACKビットの入力
    pinMode(portDATA,INPUT);                   // 入力に切り替え
 
    SCK_H();                    // ↑edge
    if( digitalRead(portDATA) ) {      // read DATA
        // 1
        rdat = 1;               // NOACK
    } else {
        // 0
        rdat = 0;               // ACK
    }
    SCK_L();
 
    return rdat;
}

//バイト・データ受信
byte SHTRead(byte ack) {
    byte i, dat;
 
    dat = 0;
    pinMode(portDATA,INPUT);                  // 入力に切り替え
 
    // DATAビットの入力
    for(i = 0; i < 8; i++) {
        dat <<= 1;                                         // 次のビット
        SCK_H();                                           // ↑edge
        if( digitalRead(portDATA) ) {          // read DATA
            // "1"
            dat |= 1;
        }
        SCK_L();
    }

    
    // ACKビットの出力
    if(!ack) {
        pinMode(portDATA,OUTPUT);               // 出力に切り替え
        DATA_L();
    }
    SCK_H();
    SCK_L();                    // ↓edge

    return dat;
}

float CalcTMP(const unsigned int ADValTmp){
	float d1;
	
	if( 		sensorVDD > 4.5 ){	//	vdd =   5.0;
		d1  = -40.1;
	}else if(	sensorVDD > 3.7 ){	//	vdd =   4.0;
		d1  = -39.8;
	}else if(	sensorVDD > 3.2 ){	//	vdd =   3.5;
		d1  = -39.7;
	}else if(	sensorVDD > 2.7 ){	//	vdd =   3.0;
		d1  = -39.6;
	}else {							//	vdd =   2.5;
		d1  = -39.4;
	}
	
	return( d1 + 0.01 * (float)ADValTmp );
}

float CalcHR(const unsigned int ADValHR , const float temp) {
	float hum = -1.5955e-6 * (float)ADValHR * (float)ADValHR;
	hum += 0.0367 * (float)ADValHR;
	hum -= 2.0468;
	hum += (temp - 25) * (0.01 + 0.00008 * (float)ADValHR );
	return( hum );
}

void HS15Init(void){
	pinMode(HS15_P_OUT, OUTPUT);   				//出力ポート動作へ
	digitalWrite(HS15_P_OUT, LOW);
}

int HS15GetRs(const unsigned int ChgTime)
{
	int j;

	/* 測定準備。C放電。*/
	pinMode(HS15_P_OUT, OUTPUT);   				// 出力ポート動作へ
	digitalWrite(HS15_P_OUT, LOW);
	for(j=0;j<20;j++){							// 200mSec又は放電完了まち
		delay( 10 );							// 10mSec
		if( analogRead(HS15_P_IN) < 5 )break;	// 放電完了をチェックし、ＯＫならば打ち切る
	}
	
	/* 測定 */
	noInterrupts();
	if( ChgTime == 5 ){							//ChgTimeはチャージ時間の半分の値
		digitalWrite(HS15_P_OUT, HIGH);
		delayMicroseconds( 10 );				// 充電 10us
		pinMode(HS15_P_OUT, INPUT);   			//ハイインピーダンスへ(入力ポート)
		digitalWrite(HS15_P_OUT, LOW);			//プルアップOFF
		j = analogRead(HS15_P_IN);				//A/D 10bit
	}else if( ChgTime == 50 ){
		digitalWrite(HS15_P_OUT, HIGH);
		delayMicroseconds( 100 );				// 充電 100us
		pinMode(HS15_P_OUT, INPUT);   			//ハイインピーダンスへ(入力ポート)
		digitalWrite(HS15_P_OUT, LOW);			//プルアップOFF
		j = analogRead(HS15_P_IN);				//A/D 10bit
	}else if( ChgTime == 500 ){
		digitalWrite(HS15_P_OUT, HIGH);
		delayMicroseconds( 1000 );				// 充電 1ms
		pinMode(HS15_P_OUT, INPUT);   			//ハイインピーダンスへ(入力ポート)
		digitalWrite(HS15_P_OUT, LOW);			//プルアップOFF
		j = analogRead(HS15_P_IN);				//A/D 10bit
	}else if( ChgTime == 5000 ){
		digitalWrite(HS15_P_OUT, HIGH);
		delay( 10 );							// 充電 10ms
		pinMode(HS15_P_OUT, INPUT);   			//ハイインピーダンスへ(入力ポート)
		digitalWrite(HS15_P_OUT, LOW);			//プルアップOFF
		j = analogRead(HS15_P_IN);				//A/D 10bit
	}else{
		digitalWrite(HS15_P_OUT, HIGH);
		delay( 100 );							// 充電 100ms
		pinMode(HS15_P_OUT, INPUT);   			//ハイインピーダンスへ(入力ポート)
		digitalWrite(HS15_P_OUT, LOW);			//プルアップOFF
		j = analogRead(HS15_P_IN);				//A/D 10bit
	}
	interrupts();
	
	pinMode(HS15_P_OUT, OUTPUT);   				//出力ポート動作へ
	digitalWrite(HS15_P_OUT, LOW);				//次の動作のために放電実行、放置
	return( j );								//AD値を応答
}

float HS15Read( const float temp ){
	int i,j;
	unsigned int t;
	float Rs;									// HS15の抵抗値（対数）
	float Rh;									// 算出した湿度

	t=5;										// 最初は10uSec充電での測定から開始する
	for(j=0; j<5; j++){
		i = HS15GetRs(t);
		Rs = -1.0e-5 / 2.2e-8 / log( 1 - (float)i /1023 );	// 抵抗値算出
		Rs = -40 * log10( Rs / 3160 );						// 対数抵抗値へ変換
		if( Rs > 0 ){
			Rs = 200 - (float)j * 40 + Rs;
			break;								// forを抜ける
		}
		t *=10;									// 10倍のレンジでリトライ
	}
	//	温度補正(元のソースリストのまま使用)
	Rs += (Rs*35-13056) /64 *temp/64;			// ※特性は直線的でないので、
	if(Rs < 0 ) Rs= 0;							// 単純にシフトするだけでは誤差が大きい
	//	リニアリティ補正(近似式：元のソースリストのまま使用)
	Rh = (5*Rs /16 *Rs + 44* Rs )/256 +20;
	return( Rh );
}

void setup(void){
	byte i;

	for( i = 0 ; i<3 ; i++ ){
		HS15_P_OUT = i+2;
		HS15_P_IN  = i;
		pinMode(HS15_P_OUT, OUTPUT);   			//出力ポート動作へ
		digitalWrite(HS15_P_OUT, LOW);
	}
	Serial.begin(9600);
	Serial.println("Hello!");
	while( SD.begin() == false ){
		Serial.println("SD ERROR");
		delay(3000);
	}
}

void loop(void){
	float hum,temp;
	unsigned int wdat;
	byte bdat_H, bdat_L, bdat_CRC;
	File file;
	char filename[] = "hs15p4_1.csv";
	char Str[16];
	
	float hs15[3];
	byte port;
	byte i;

	SHTInit();

	SHTTSSeq();							// TSシーケンス発行
	SHTWrite(0b00000011);				// measure temprature コマンド送信
	while( digitalRead(portDATA) );		// A-D変換完了を待つ
	bdat_H = SHTRead(0);				// read high byte(ACK)
	bdat_L = SHTRead(0);				// read low byte(ACK)
	bdat_CRC = SHTRead(1);				// read CRC(NOACK)

	wdat = ((unsigned int)bdat_H << 8) | bdat_L;                 // 16ビットに合成
	temp = CalcTMP(wdat);
	
	SHTTSSeq();							// TSシーケンス発行
	SHTWrite(0b00000101);				// measure humidity コマンド送信
	while( digitalRead(portDATA) );		// A-D変換完了を待つ
	bdat_H = SHTRead(0);				// read high byte(ACK)
	bdat_L = SHTRead(0);				// read low byte(ACK)
	bdat_CRC = SHTRead(1);				// read CRC(NOACK)

	wdat = ((unsigned int)bdat_H << 8) | bdat_L;                 // 16ビットに合成
	hum  = CalcHR(wdat,temp);
	
	Serial.print("SHT TEMP = ");
	Serial.print( temp , 2 );
	Serial.print(" degC\n");
	Serial.print("SHT HUM  = ");
	Serial.print( hum , 2 );
	Serial.print(" %\n");

	for( i = 0 ; i<3 ; i++ ){
		
		HS15_P_OUT = i+2;
		HS15_P_IN  = i;

		/* HS-15P */
		HS15Init();
		hs15[i] = HS15Read( temp );
		
		/* シリアル出力 */
		Serial.print("HS15 [");
		Serial.print( i, DEC);
		Serial.print("] = ");
		Serial.print( hs15[i] , 1 );
		Serial.print(" %\n");
	}

	Serial.print("writing SD ... ");
	if( (file = SD.open( filename, FILE_WRITE )) != false ){
		file.print( temp , 2 );
		file.print( ", " );
		file.print( hum , 2 );
		file.print( ", " );
		for( i = 0 ; i<3 ; i++ ){
			file.print( hs15[i] , 2 );
			if( i != 3 ) file.print( ", " );
		}
		file.print( "\n" );
		file.close();
		Serial.println( "done" );
	} else Serial.println("failed file open");
	delay(30000);
	delay(30000);
}
